/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>
#include <math.h>

#define PN_CLOCK_OUT 0
#define PN_BAR_RESET_OUT 1
#define PN_TEMPO_OUT 2
#define PN_TRANSPORT_OUT 3

#define DEFAULT_TEMPO 10 // 100 BPM

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
,pulse_count(0.f)
,pulse_count_int(0)
,m_midi_clock(-1)
,quarter_note_count(-1)
,quarters_per_bar(4)
,error(0.f)
,bpm_out(this,PN_TEMPO_OUT)
,clock_out(this,PN_CLOCK_OUT)
,bar_reset_out(this,PN_BAR_RESET_OUT)
,transport_out(this,PN_TRANSPORT_OUT)
{
}

//#define CLOCKS_PER_BEAT 24.0

void Module::sub_process(long buffer_offset, long sampleFrames )
{
	float samples_per_beat;
	VstTimeInfo *ti = 0;

	int samples_remain = sampleFrames;
	int buffer_position = buffer_offset;

	while(true)
	{
		int to_do = samples_remain;
		if( pulse_count_int < samples_remain )
		{
			to_do = pulse_count_int;
		}

		bpm_out.process( buffer_position, to_do );
		clock_out.process( buffer_position, to_do );
		bar_reset_out.process( buffer_position, to_do );
		transport_out.process( buffer_position, to_do );

		samples_remain -= to_do;
		pulse_count_int -= to_do;

		if( samples_remain == 0 )
			return;

		buffer_position += to_do;

		assert( pulse_count_int == 0 );
		assert( pulse_count <= 0.f );
		{
			int samples_processed_so_far = sampleFrames - samples_remain;
			unsigned int cur_sample_clock = SampleClock() + samples_processed_so_far;

			float new_bar_reset = 0.f;
			if( clock_out.GetValue() == 0.f )
			{
				short safe_pulses_per_beat = pulses_per_beat;
				if( safe_pulses_per_beat < 1 )
					safe_pulses_per_beat = 1;

				// keep track of position within beat (24 clock/beat)
				m_midi_clock++;
				m_midi_clock %= safe_pulses_per_beat;

				if( m_midi_clock == 0) // start of quarter note
				{
					quarter_note_count++;
					quarter_note_count %= quarters_per_bar;
					if( quarter_note_count == 0) // start of bar
					{
						new_bar_reset = 0.5f;
					}
				}

				clock_out.SetValue( cur_sample_clock, 0.5f );
			}
			else
			{
				clock_out.SetValue( cur_sample_clock, 0.f );
			}

			bar_reset_out.SetValue( cur_sample_clock, new_bar_reset );

			if( ti == 0 ) // only do this a max of once per call
			{
				// add on time since block start
				ti = (VstTimeInfo *) CallHost(seaudioMasterGetTime, 0, kVstTempoValid|kVstPpqPosValid, 0 );
				if( ti != 0 )
				{
					short safe_pulses_per_beat = pulses_per_beat;
					if( safe_pulses_per_beat < 1 )
						safe_pulses_per_beat = 1;

					// calculate how many samples per beat/clock
					double samples_to_beats = ti->tempo / ( sampleRate * 60.0 ); // converts secs to beats (quarter notes)
					double samples_to_clocks = samples_to_beats * safe_pulses_per_beat;

					//int samples_processed_so_far = sampleFrames - samples_remain;//s;
					// half a midi clock (for on/off transition)
					samples_per_beat = 0.5f / (float)samples_to_clocks;

					float new_transport_running = 0.f;

					// lock to host if transport playing, else just freewheel
					if( (ti->flags & kVstTransportPlaying) != 0 )
					{
						new_transport_running = 1.f;

						double extra_samples = buffer_offset + (double) samples_processed_so_far;
						double host_total_beats = ti->ppqPos + extra_samples * samples_to_beats;
						quarter_note_count = (int) host_total_beats;

						float beat_frac = (float) m_midi_clock;

						// this routine called twice per midi clock (leading edge/trailing edge)
						// on trailing edge, we're half way between midi clocks
						if( clock_out.GetValue() == 0.f )//outval == 0.f )
						{
							beat_frac += 0.5;
						}

						// add small fractional ammount due to pulse not always
						// falling on an exact sampleframe
						beat_frac -= 0.5f * pulse_count/samples_per_beat;

						// convert from MIDI clocks to beats
						beat_frac = beat_frac / (float) safe_pulses_per_beat;

						float host_beat_frac = (float)(host_total_beats - floor( host_total_beats ));
						
						// calc error as fraction of beat
						error = host_beat_frac - beat_frac;
						if( error < -0.5 )
							error += 1.0;
						if( error > 0.5 )
							error -= 1.0;

						// convert error back to samples
						float correction = error * 2.f * samples_per_beat * safe_pulses_per_beat;

						// jump count a fraction to catch up with host
						pulse_count -= correction; //samples_per_beat * error * 0.8f;

//#ifdef _DEBUG
//						getPin(PN_OUTPUT5)->TransmitStatusChange( SampleClock() + samples_processed_so_far, ST_ONE_OFF );
//#endif
					}
					
					float new_tempo = (float) ti->tempo * 0.1f;

					// avoid divide-by-zero errors when host don't provide tempo info
					if( new_tempo == 0.f )
						new_tempo = DEFAULT_TEMPO;

					// note: these are not sample-accurate. Being Polled, they always lag actual event.
					bpm_out.SetValue      ( cur_sample_clock, new_tempo  );
					transport_out.SetValue( cur_sample_clock, new_transport_running );
				}
			}

			// reset pulse count
			pulse_count += samples_per_beat;

			// can't skip a pulse completely, that would lose sync. Enforce minimum duration.
			if( pulse_count < 2.f )
			{
				pulse_count = 2.f;
			}

			// gracefully fail in buggy mixcraft. Where pulse count overfloaws to #inf.
			if( !( pulse_count < 1000000.f ) )
			{
				pulse_count = 44100.f;
			}

			pulse_count_int = (int) ceil(pulse_count);

			// pre-calculate pulse count at next clock
			pulse_count -= pulse_count_int;
		}
/*
#ifdef _DEBUG
		*out4++ = error;
#endif
*/
	}
}

void Module::open()
{
	SET_PROCESS_FUNC(Module::sub_process);
	SEModule_base::open();
	bpm_out.SetValue( SampleClock(), DEFAULT_TEMPO );
}

// describe the module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	// describe the plugin, this is the name the end-user will see.
	properties->name = "BPM Clock2";

	// return a unique string 32 characters max
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug. No two plugs may have the same id.
	properties->id = "Synthedit BPM Clock V2";

	// Info, may include Author, Web page whatever
	properties->about = "by Jeff M";
	return true;
}

// describe the pins (plugs)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )
	{
	case 0:
		properties->name				= "Pulse Out";
		properties->variable_address	= clock_out.PointerAddress();//&output1_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 1:
		properties->name				= "Bar Start";
		properties->variable_address	= bar_reset_out.PointerAddress();//&output4_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 2:
		properties->name				= "Tempo Out";
		properties->variable_address	= bpm_out.PointerAddress();
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 3:
		properties->name				= "Transport Run";
		properties->variable_address	= transport_out.PointerAddress();//&output3_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 4:
		properties->name				= "Pulse Divide";
		properties->variable_address	= &pulses_per_beat;
		properties->datatype_extra		= "64=64,32=32,24=24,16=16,12=12,8=8,6=6,4=4,3=3,2=2,1=1";
		properties->default_value		= "4";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
/*
#ifdef _DEBUG
	case 4:
		properties->name				= "Diagnostic";
		properties->variable_address	= &output4_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
#endif
*/
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}
